@@ -7,12 +7,19 @@ use std::sync::Mutex;
 
 pub use turbocharger_impl::{backend, server_only, wasm_only};
 
+#[wasm_only]
+#[cfg(feature = "svelte")]
+#[wasm_bindgen(typescript_custom_section)]
+const Subscriber: &'static str = r#"
+import { Subscriber } from "svelte/store";
+"#;
+
 #[doc(hidden)]
 pub use {bincode, futures, serde};
 
 #[server_only]
 #[doc(hidden)]
-pub use {async_stream, async_trait::async_trait, typetag};
+pub use {async_stream, async_trait::async_trait, stream_cancel, typetag};
 
 #[cfg(target_arch = "wasm32")]
 #[doc(hidden)]
@@ -23,19 +30,28 @@ pub use js_sys;
 #[typetag::serde]
 #[async_trait]
 pub trait RPC: Send + Sync {
- async fn execute(&self, sender: Box<dyn Fn(Vec<u8>) + Send>);
+ async fn execute(
+  &self,
+  sender: Box<dyn Fn(Vec<u8>) + Send>,
+  tripwire: Option<stream_cancel::Tripwire>,
+  remote_addr: Option<std::net::SocketAddr>,
+  user_agent: Option<String>,
+ );
+ fn txid(&self) -> i64;
 }
 
 struct Globals {
  #[cfg_attr(not(target_arch = "wasm32"), allow(dead_code))]
  channel_tx: Option<futures::channel::mpsc::UnboundedSender<Vec<u8>>>,
+ #[cfg_attr(not(target_arch = "wasm32"), allow(dead_code))]
+ socket_url: Option<String>,
  next_txid: i64,
  senders: std::collections::HashMap<i64, futures::channel::mpsc::UnboundedSender<Vec<u8>>>,
 }
 
 impl Default for Globals {
  fn default() -> Self {
-  Self { channel_tx: None, next_txid: 256, senders: Default::default() }
+  Self { socket_url: None, channel_tx: None, next_txid: 256, senders: Default::default() }
  }
 }
 
@@ -84,95 +100,17 @@ extern "C" {
  fn log(s: &str);
 }
 
-/// Returns a combined `warp` route for serving both the Turbocharger WebSocket and the static frontend files.
-#[server_only]
-pub fn warp_routes<A: 'static + rust_embed::RustEmbed>(
- asset: A,
-) -> warp::filters::BoxedFilter<(impl warp::Reply,)> {
- use warp::Filter;
- warp_socket_route().or(warp_rust_embed_route(asset)).boxed()
-}
+#[cfg(not(target_arch = "wasm32"))]
+#[cfg(feature = "axum_server")]
+mod axum_server;
 
-/// Returns a `warp` route for serving the Turbocharger WebSocket.
-#[server_only]
-pub fn warp_socket_route() -> warp::filters::BoxedFilter<(impl warp::Reply,)> {
- use warp::Filter;
- warp::path("turbocharger_socket")
-  .and(warp::ws())
-  .map(|ws: warp::ws::Ws| ws.on_upgrade(accept_connection))
-  .boxed()
-}
+#[cfg(not(target_arch = "wasm32"))]
+#[cfg(feature = "axum_server")]
+pub use axum_server::serve;
 
-/// Returns a `warp` route for serving static frontend files sourced from the `rust_embed` crate.
-#[server_only]
-pub fn warp_rust_embed_route<A: rust_embed::RustEmbed>(
- _asset: A,
-) -> warp::filters::BoxedFilter<(impl warp::Reply,)> {
- use warp::Filter;
- warp::path::full()
-  .map(|path: warp::path::FullPath| {
-   let path = match path.as_str().trim_start_matches('/') {
-    "" => "index.html",
-    path => path,
-   };
-   match A::get(path) {
-    None => warp::http::Response::builder().status(404).body("404 not found!".into()).unwrap(),
-    Some(asset) => {
-     let mime = mime_guess::from_path(path).first_or_octet_stream();
-     let mut res = warp::reply::Response::new(asset.data.into());
-     res
-      .headers_mut()
-      .insert("content-type", warp::http::header::HeaderValue::from_str(mime.as_ref()).unwrap());
-     res
-    }
-   }
-  })
-  .boxed()
-}
-
-#[server_only]
-async fn accept_connection(ws: warp::ws::WebSocket) {
- use futures::TryFutureExt;
-
- log::debug!("accept_connection");
-
- let (mut ws_tx, mut ws_rx) = ws.split();
- let (tx, rx) = tokio::sync::mpsc::unbounded_channel();
- let mut rx = tokio_stream::wrappers::UnboundedReceiverStream::new(rx);
-
- tokio::task::spawn(async move {
-  while let Some(msg) = rx.next().await {
-   ws_tx
-    .send(msg)
-    .unwrap_or_else(|e| {
-     log::warn!("websocket send error: {}", e);
-    })
-    .await;
-  }
- });
-
- while let Some(result) = ws_rx.next().await {
-  let msg = match result {
-   Ok(msg) => msg,
-   Err(e) => {
-    log::warn!("websocket error: {}", e);
-    break;
-   }
-  };
-  let tx_clone = tx.clone();
-  tokio::task::spawn(async move {
-   let msg = msg.as_bytes();
-   if !msg.is_empty() {
-    let target_func: Box<dyn RPC> = bincode::deserialize(msg).unwrap();
-    let sender =
-     Box::new(move |response| tx_clone.send(warp::ws::Message::binary(response)).unwrap());
-    target_func.execute(sender).await;
-   }
-  });
- }
-
- log::warn!("accept_connection completed")
-}
+#[cfg(not(target_arch = "wasm32"))]
+#[cfg(all(feature = "tls", feature = "axum_server"))]
+pub use axum_server::serve_tls;
 
 #[doc(hidden)]
 pub struct _Transaction {
@@ -196,6 +134,7 @@ impl _Transaction {
  #[cfg(target_arch = "wasm32")]
  pub fn send_ws(&self, req: Vec<u8>) {
   wasm_bindgen_futures::spawn_local(async move {
+   ensure_ws_connected().await;
    let mut channel = G.lock().unwrap().channel_tx.clone().unwrap();
    channel.send(req).await.unwrap();
   });
@@ -229,11 +168,12 @@ impl Default for _Transaction {
 
 /// _Experimental._ Spawns a new Turbocharger UDP server. Future resolves when the server is ready to respond to requests.
 #[server_only]
+#[tracked::tracked]
 #[doc(hidden)]
-pub async fn spawn_udp_server(port: u16) -> Result<(), Box<dyn std::error::Error>> {
+pub async fn spawn_udp_server(port: u16) -> tracked::Result<()> {
  let socket = std::sync::Arc::new(tokio::net::UdpSocket::bind(format!("0.0.0.0:{}", port)).await?);
  log::debug!("Listening on: {}", socket.local_addr()?);
- *UDP_SOCKET.lock()? = Some(socket.clone());
+ *UDP_SOCKET.lock().unwrap() = Some(socket.clone());
 
  tokio::spawn(async move {
   loop {
@@ -257,7 +197,7 @@ pub async fn spawn_udp_server(port: u16) -> Result<(), Box<dyn std::error::Error
         send_socket_cloned.send_to(&response, peer).await.unwrap();
        });
       });
-      target_func.execute(sender).await;
+      target_func.execute(sender, None, Some(peer), Some("udp".into())).await;
      });
     }
     txid => {
@@ -272,51 +212,66 @@ pub async fn spawn_udp_server(port: u16) -> Result<(), Box<dyn std::error::Error
  Ok(())
 }
 
+#[wasm_only]
+#[wasm_bindgen]
+pub fn set_socket_url(url: String) {
+ let mut g = G.lock().unwrap();
+ g.socket_url = Some(url);
+}
+
+#[wasm_only]
+async fn ensure_ws_connected() {
+ let mut g = G.lock().unwrap();
+
+ if g.socket_url.is_none() {
+  g.socket_url = Some({
+   let location = web_sys::window().unwrap().location();
+   let protocol = match location.protocol().unwrap().as_str() {
+    "https:" => "wss:",
+    _ => "ws:",
+   };
+   format!("{}//{}/turbocharger_socket", protocol, location.host().unwrap())
+  });
+ }
+ let socket_url = g.socket_url.clone().unwrap();
+
+ if g.channel_tx.is_none() {
+  let (channel_tx, mut channel_rx) = futures::channel::mpsc::unbounded();
+  g.channel_tx = Some(channel_tx);
+
+  drop(g);
+
+  tc_console_log!("connecting to {}", socket_url);
+
+  let (_ws, wsio) = ws_stream_wasm::WsMeta::connect(socket_url, None).await.unwrap();
+
+  tc_console_log!("connected");
+
+  let (mut ws_tx, mut ws_rx) = wsio.split();
+
+  wasm_bindgen_futures::spawn_local(async move {
+   while let Some(msg) = ws_rx.next().await {
+    if let ws_stream_wasm::WsMessage::Binary(msg) = msg {
+     let txid = i64::from_le_bytes(msg[0..8].try_into().unwrap());
+     let mut sender = G.lock().unwrap().senders.get(&txid).unwrap().clone();
+     sender.send(msg).await.unwrap();
+    }
+   }
+   tc_console_log!("ws_rx ENDED");
+  });
+
+  wasm_bindgen_futures::spawn_local(async move {
+   while let Some(msg) = channel_rx.next().await {
+    ws_tx.send(ws_stream_wasm::WsMessage::Binary(msg)).await.unwrap();
+   }
+   tc_console_log!("rx ENDED");
+  });
+ }
+}
+
 #[wasm_only]
 #[wasm_bindgen(start)]
 pub async fn start() -> Result<(), JsValue> {
  console_error_panic_hook::set_once();
-
- let (channel_tx, mut channel_rx) = futures::channel::mpsc::unbounded();
- G.lock().unwrap().channel_tx = Some(channel_tx);
-
- #[cfg(turbocharger_test)]
- let socket_url = "ws://localhost:8080/turbocharger_socket";
- #[cfg(not(turbocharger_test))]
- let socket_url = {
-  let location = web_sys::window().unwrap().location();
-  let protocol = match location.protocol().unwrap().as_str() {
-   "https:" => "wss:",
-   _ => "ws:",
-  };
-  format!("{}//{}/turbocharger_socket", protocol, location.host().unwrap())
- };
-
- tc_console_log!("connecting to {}", socket_url);
-
- let (_ws, wsio) = ws_stream_wasm::WsMeta::connect(socket_url, None).await.unwrap();
-
- tc_console_log!("connected");
-
- let (mut ws_tx, mut ws_rx) = wsio.split();
-
- wasm_bindgen_futures::spawn_local(async move {
-  while let Some(msg) = ws_rx.next().await {
-   if let ws_stream_wasm::WsMessage::Binary(msg) = msg {
-    let txid = i64::from_le_bytes(msg[0..8].try_into().unwrap());
-    let mut sender = G.lock().unwrap().senders.get(&txid).unwrap().clone();
-    sender.send(msg).await.unwrap();
-   }
-  }
-  tc_console_log!("ws_rx ENDED");
- });
-
- wasm_bindgen_futures::spawn_local(async move {
-  while let Some(msg) = channel_rx.next().await {
-   ws_tx.send(ws_stream_wasm::WsMessage::Binary(msg)).await.unwrap();
-  }
-  tc_console_log!("rx ENDED");
- });
-
  Ok(())
 }
